import xlrd
from qingtest.utility import Excel, data2dict
from qingtest.config import header
from qingtest.globals import g


def testsuite_format(data):
    '''
    将元素为 dict 的 list，处理为 testcase 的 list
    testcase 的格式：
    {
        'id': 'Login_001',  #用例编号
        'title': 'Login OK',  #用例标题
        'condition': '',  #前置条件
        'designer': 'Leo',  #设计者
        'flag': '',  #自动化标记
        'result': '',  #用例结果
        'remark': '',  #备注
        'steps':
            [
                {
                'no': 1,  #测试步骤
                'keyword': '输入',
                'page': '产品管系统登录页',
                'element': '用户名',
                'data': 'user1',  #测试数据
                'output': '',  #输出数据
                'score': '',  #测试结果
                'remark': ''  #备注
                },
                {……}
                ……
            ]
    }
    '''
    testsuite = []
    testcase = {'testsuite': '', 'no': 0}
    data = data2dict(data)

    for d in data:
        # 如果用例编号不为空，则为新的用例
        if d['id'].strip():
            # 如果 testcase[id] 非空，则添加到 testsuite 里，并重新初始化 testcase
            if testcase.get('id'):
                testsuite.append(testcase)
                testcase = {'testsuite': '', 'no': 0}
            for key in ('id', 'title', 'condition', 'designer', 'flag', 'result', 'remark'):
                testcase[key] = d[key]
            if '#' in d['id']:
                testcase['set'] = d['id'].split('#')[0]
                testcase['flag'] = 'N'
            testcase['priority'] = d['priority'] if d['priority'] else 'M'
            testcase['steps'] = []
        # 如果测试步骤不为空，则为有效步骤，否则用例解析结束
        no = str(d['step']).strip()
        if no:
            step = {}
            step['control'] = ''
            if no[0] in ('^', '>', '<', '#'):
                step['control'] = no[0]
                step['no'] = no
            else:
                step['no'] = str(int(d['step']))
            for key in ('keyword', 'page', 'element', 'data', 'expected', 'output', 'score', 'remark'):
                step[key] = d.get(key, '')

            # 仅作为测试结果输出时，保持原样
            step['_keyword'] = d['keyword']
            step['_element'] = d['element']
            step['_data'] = d['data']
            step['vdata'] = d.get('data', '')
            step['_expected'] = d.get('expected', '')
            step['_output'] = d.get('output', '')
            testcase['steps'].append(step)
    if testcase:
        testsuite.append(testcase)
    return testsuite


def testsuite_from_excel(file_name, sheet_name):
    d = Excel(file_name)
    return testsuite_format(data2dict(d.read(sheet_name)))


def testsuite2data(data):
    # result = [list(header.values())]
    result = [[g.header_custom[key.lower()] for key in header.values()]]
    for d in data:
        s = d['steps'][0]  # 第一步和用例标题同一行
        testcase = [d['id'], d['title'], d['condition'], s['no'], s['_keyword'], s['page'], s['_element'],
                    s['vdata'], s['_output'], d['priority'], d['designer'], d['flag'], s['score'], d['result'], s['remark']]
        if g.header_custom['expected']:
            testcase.insert(8, s['_expected'])
        result.append(testcase)
        for s in d['steps'][1:]:
            step = ['', '', '', s['no'], s['_keyword'], s['page'], s['_element'],
                    s['vdata'], s['_output'], '', '', '', s['score'], '', s['remark']]
            if g.header_custom['expected']:
                step.insert(8, s['_expected'])
            result.append(step)
    return result


def testsuite2report(data):
    report = []
    for case in data:
        if case['condition'] in ('BASE', 'SETUP') or case['flag'] != 'N':
            for step in case['steps']:
                step['keyword'] = step.pop('_keyword')
                step['element'] = step.pop('_element')
                step['data'] = str(step.pop('vdata'))
                step['expected'] = step.pop('_expected')
                step['output'] = step.pop('_output')
            report.append(case)
    return report



from selenium.webdriver.common.keys import Keys
from pathlib import Path
import xlrd
import xlsxwriter
from openpyxl import Workbook, load_workbook
import csv
import re
import json
import time
import random
from qingtest.config import header
from qingtest.globals import g


path = Path('lib')
if path.is_dir():
    from lib import *
else:
    from qingtest.lib import *


class Excel:
    def __init__(self, file_name, mode='r'):
        if mode == 'r':
            self.workbook = xlrd.open_workbook(file_name)
        elif mode == 'w':
            self.workbook = xlsxwriter.Workbook(file_name)
        else:
            raise Exception(
                'Error: init Excel class with error mode: %s' % mode)

    def get_sheet(self, sheet_name):
        names = []
        if isinstance(sheet_name, str):
            if sheet_name.endswith('*'):
                for name in self.workbook.sheet_names():
                    if sheet_name[:-1] in name:
                        names.append(name)
            else:
                names.append(sheet_name)
        elif isinstance(sheet_name, list):
            names = sheet_name
        else:
            raise Exception('Error: invalidity sheet_name: %s' % sheet_name)

        return names

    def read(self, sheet_name):
        '''
        sheet_name:Excel 中标签页名称
        return：[[],[]……]
        '''
        sheet = self.workbook.sheet_by_name(sheet_name)
        nrows = sheet.nrows
        data = []
        for i in range(nrows):
            data.append(sheet.row_values(i))
        return data

    def write(self, data, sheet_name):
        sheet = self.workbook.add_worksheet(sheet_name)

        red = self.workbook.add_format({'bg_color': 'red', 'color': 'white'})
        gray = self.workbook.add_format({'bg_color': 'gray', 'color': 'white'})
        green = self.workbook.add_format(
            {'bg_color': 'green', 'color': 'white'})
        blue = self.workbook.add_format({'bg_color': 'blue', 'color': 'white'})
        orange = self.workbook.add_format(
            {'bg_color': 'orange', 'color': 'white'})
        for i in range(len(data)):
            for j in range(len(data[i])):
                if str(data[i][j]) == 'failure':
                    sheet.write(i, j, str(data[i][j]), red)
                elif str(data[i][j]) == 'NO':
                    sheet.write(i, j, str(data[i][j]), gray)
                elif str(data[i][j]) == 'blocked':
                    sheet.write(i, j, str(data[i][j]), orange)
                elif str(data[i][j]) == 'skipped':
                    sheet.write(i, j, str(data[i][j]), blue)
                elif str(data[i][j]) == 'success':
                    sheet.write(i, j, str(data[i][j]), green)
                else:
                    sheet.write(i, j, data[i][j])

    def close(self):
        self.workbook.close()


class Excel_01():
    def __init__(self, file_name, mode='r'):
        if mode == 'r':
            self.workbook = load_workbook(file_name)
        elif mode == 'w':
            self.workbook = Workbook() #.Workbook(file_name)
        else:
            raise Exception(
                'Error: init Excel class with error mode: %s' % mode)

    def get_sheet(self, sheet_name):
        names = []
        if isinstance(sheet_name, str):
            if sheet_name.endswith('*'):
                for name in self.workbook.sheet_names():
                    if sheet_name[:-1] in name:
                        names.append(name)
            else:
                names.append(sheet_name)
        elif isinstance(sheet_name, list):
            names = sheet_name
        else:
            raise Exception('Error: invalidity sheet_name: %s' % sheet_name)

        return names

    def read(self, sheet_name):
        '''
        sheet_name:Excel 中标签页名称
        return：[[],[]……]
        '''
        sheet = self.workbook.sheet_by_name(sheet_name)
        nrows = sheet.nrows
        data = []
        for i in range(nrows):
            data.append(sheet.row_values(i))
        return data

    def write(self, data, sheet_name):
        sheet = self.workbook.add_worksheet(sheet_name)

        red = self.workbook.add_format({'bg_color': 'red', 'color': 'white'})
        gray = self.workbook.add_format({'bg_color': 'gray', 'color': 'white'})
        green = self.workbook.add_format(
            {'bg_color': 'green', 'color': 'white'})
        blue = self.workbook.add_format({'bg_color': 'blue', 'color': 'white'})
        orange = self.workbook.add_format(
            {'bg_color': 'orange', 'color': 'white'})
        for i in range(len(data)):
            for j in range(len(data[i])):
                if str(data[i][j]) == 'failure':
                    sheet.write(i, j, str(data[i][j]), red)
                elif str(data[i][j]) == 'NO':
                    sheet.write(i, j, str(data[i][j]), gray)
                elif str(data[i][j]) == 'blocked':
                    sheet.write(i, j, str(data[i][j]), orange)
                elif str(data[i][j]) == 'skipped':
                    sheet.write(i, j, str(data[i][j]), blue)
                elif str(data[i][j]) == 'success':
                    sheet.write(i, j, str(data[i][j]), green)
                else:
                    sheet.write(i, j, data[i][j])

    def close(self):
        self.workbook.close()


def data2dict(data):
    # def list_list2list_dict(data):
    '''
    把带头标题的二维数组，转换成以标题为 key 的 dict  的 list
    '''
    list_dict_data = []
    key = []
    g.header_custom = {}  # 用户自定义的标题
    for d in data[0]:
        k = d.strip().split('#')[0]
        # 如果为中文，则替换成英文
        h = header.get(k, k).lower()
        key.append(h)
        g.header_custom[h] = d.strip()

    if not g.header_custom.get('expected'):
        g.header_custom['expected'] = ''

    for d in data[1:]:
        dict_data = {}
        for i in range(len(key)):
            if isinstance(d[i], str):
                dict_data[key[i]] = str(d[i]).strip()
            else:
                dict_data[key[i]] = d[i]
        list_dict_data.append(dict_data)
    return list_dict_data


def replace_dict(data):
    # 变量替换
    for key in data:
        data[key] = replace(data[key])


def replace_list(data):
    # 变量替换
    for i in range(len(data)):
        data[i] = replace(data[i])


def replace_old(data):
    # 正则匹配出 data 中所有 <> 中的变量，返回列表
    keys = re.findall(r'<(.*?)>', data)
    for k in keys:
        # 正则匹配出 k 中的 + - ** * // / % , ( ) 返回列表
        values = re.split(r'(\+|-|\*\*|\*|//|/|%|,|\(|\)|\'|\")', k)
        for j, v in enumerate(values):
            #切片操作处理，正则匹配出 [] 中内容
            s = v.split('[', 1)
            index = ''
            if len(s) == 2:
                v = s[0]
                index = '[' + s[1]

            if v in g.var:
                # 如果在 g.var 中是 list
                if isinstance(g.var[v], list):
                    if index:
                        # list 切片取值（值应该是动态赋值的变量，如自定义脚本的返回值）
                        values[j] = eval('g.var[v]' + index)
                    else:
                        if len(g.var[v]) == 1:
                            values[j] = g.var[v][0]
                            g.var['_last_'] = True
                        else:
                            values[j] = g.var[v].pop(0)
                elif isinstance(g.var[v], dict) and index:
                    # 如果是 dict 取键值
                    values[j] = eval('g.var[v]' + index)
                else:
                    # 如果在 g.var 中是值，则直接赋值
                    values[j] = g.var[v]
                    if index:
                        values[j] = eval('g.var[v]' + index)
            # 如果值不在 g.var，且只有一个元素，则尝试 eval，比如<False>,<True>,<1>,<9.999>
            elif len(values) == 1:
                try:
                    values[j] = eval(values[j])
                except:
                    pass

        # 如果 values 长度大于 1，说明有算术运算符，则用 eval 运算
        # 注意，先要把元素中的数字变为字符串
        if len(values) > 1:
            values = eval(''.join([str(x) for x in values]))
        # 如果 values 长度为 1，则直接赋值，注意此值可能是数字
        else:
            values = values[0]
        # 如果 data 就是一个 <>，如 data = '<a+1>',则直接赋值为 values，此值可能是数字
        if data == '<' + keys[0] + '>':
            data = values
            # 如果有键盘操作，则需要 eval 处理
            if isinstance(data, str) and 'Keys.' in data:
                data = eval(data)
        # 否则需要替换，此时变量强制转换为为字符串
        else:
            data = data.replace('<' + k + '>', str(values))
    return data


def replace(data):
    
    if data.startswith("'''") and data.endswith("'''"):
        return data[3:-3]

    left_angle = 'dsfw4rwfdfstg43'
    right_angle = '3dsdtgsgt43trfdf'
    left_delimiter = '<'
    right_delimiter = '>'
    data = data.replace(r'\<', left_angle).replace(r'\>', right_angle)
    
    if '<<' in data and '>>' in data:
        left_delimiter = '<<'
        right_delimiter = '>>'           
        
    # 正则匹配出 data 中所有 <> 中的变量，返回列表
    keys = re.findall(r'%s' %(left_delimiter+'(.*?)'+right_delimiter), data)
    _vars = {}

    for k in keys:
        k = k.replace(left_angle, '<').replace(right_angle, '>')
        # 正则匹配出 k 中的 + - ** * // / % , ( ) 返回列表
        values = re.split(r'(\+|-|\*\*|\*|//|/|%|,|\(|\))', k)
        for v in values:
            #切片操作处理，正则匹配出 [] 中内容
            s = v.split('[', 1)
            index = ''
            if len(s) == 2:
                v = s[0]
                index = '[' + s[1]

            if v not in _vars:
                if v in g.var:
                    # 如果在 g.var 中是值，则直接赋值
                    _vars[v] = g.var[v]
                elif v in g.test_data:
                        # 是测试数据文件中的值
                        if len(g.test_data[v]) == 0:    # 为空 list 报错
                            raise Exception('The key:%s is no value in data csv' %v)
                        elif len(g.test_data[v]) == 1:  # 最后一个 list 元素，赋值
                            _vars[v] = g.test_data[v][0]
                            g.var['_last_'] = True  # 处理循环次数为 N 的情况，此标志标明循环到最后一个变量
                        else:                           # 大于 1 个元素，pop 赋值
                            _vars[v] = g.test_data[v].pop(0)

        try:
            value = eval(k, globals(), _vars)    
        except NameError:                  
            value = left_delimiter+ k + right_delimiter              

        if data == left_delimiter+ keys[0] + right_delimiter:
            data = value
        # 否则需要替换，此时变量强制转换为为字符串
        else:
                data = data.replace(left_delimiter + k + right_delimiter, str(value))

    if isinstance(data, str):            
        data = data.replace(left_angle, '<').replace(right_angle, '>')
                        
    return data


def test_replace():
    g.var = {'a': 1, 'b': 'B'}
    for d in ('<a+1>', '<b>', 'abc<a>', 'abc<a+1>', '<a*(8+4)/2//3>', '<u.td(-3)>'):
        print('data:%s' % d)
        data = replace(d)
        print(repr(data))


def read_csv(csv_file, encoding=None):
    data = []
    with open(csv_file, encoding=encoding) as f:
        reader = csv.reader(f)
        for line in reader:
            data.append(line)
    return data


def write_csv(csv_file, data, encoding=None):
    with open(csv_file, 'w', encoding=encoding, newline='') as f:
        writer = csv.writer(f)
        writer.writerows(data)


def get_record(data_file):
    encoding = None
    try:
        data = read_csv(data_file, encoding='utf-8')
        encoding = 'utf-8'
    except:
        data = read_csv(data_file)

    def read_data():
        num = len(data[0])-1 if flag else len(data[0])
        for i in range(num):
            if d[i]:
                k = data[0][i]
                if record.get(k, None):
                    # if isinstance(record[k], str):
                    #    record[k] = [record[k]]
                    record[k].append(d[i])
                else:
                    record[k] = [d[i]]
                if record[k][-1] == '&quot;':  # 空字符转换
                    record[k][-1] = ''

    record = {}
    flag = False
    if data[0][-1].lower() == 'flag':
        flag = True

    for d in data[1:]:
        if not flag:
            read_data()
        elif d[-1] == 'N':
            read_data()    
        elif d[-1] != 'Y':
            read_data()
            d[-1] = 'Y'
            write_csv(data_file, data, encoding=encoding)
            break
    return record


def str2int(s):
    s = str(s).replace(',', '').split('.', 1)
    if len(s) == 2:
        dot = s[-1]
        assert int(dot) == 0
    return int(s[0])


def zero(s):
    if s and s[-1] == '0':
        s = s[:-1]
        s = zero(s)
    return s


def str2float(s, n=None):
    s = str(s).replace(',', '')
    number = s.split('.', 1)
    if n:
        f = float(s)
        return round(f, n), n

    dot = '0'
    if len(number) == 2:
        dot = number[-1]
        dot = zero(dot)
    f = float(number[0] + '.' + dot)

    return round(f, len(dot)), len(dot)


def f(v, e, n=2):
    '''
    判断2个 float 数值是否相同，类型可以是 str 或 float 
    v: 实际值，如：12.345, '1234.56', '1,234.5600' 
    e: 预期结果, 示例值同 v
    n: 小数点精确位数
    '''
    v = str(v).replace(',', '')
    e = str(e).replace(',', '') 
    _v = round(float(v), n)
    _e = round(float(e), n)
    assert round(_v, n) == round(_e, n)


def mkdir(p):
    path = Path(p)
    if not path.is_dir():
        path.mkdir()


def json2dict(s):
    if isinstance(s, dict):
        return s
    s = str(s)
    d = {}
    try:
        d = json.loads(s)
    except:
        try:
            d = eval(s)
        except:
            s = s.replace('true', 'True').replace('false', 'False').replace(
                'null', 'None').replace('none', 'None')
            d = eval(s)
    return d


def compare(data, real):
    if isinstance(data, str):

        if data.startswith('#'):
            assert data[1:] != str(real)
            return
        elif data.startswith(':'):
            exec('v=real;'+data[1:])
            return
                    
        assert isinstance(real, str)

        if data.startswith('*'):
            assert data[1:] in real
            return
        elif data.startswith('^'):
            assert real.startswith(data[1:])
            return
        elif data.startswith('$'):
            assert real.endswith(data[1:])
            return

        elif data.startswith('\\'):
            data = data[1:]

        assert data == real

    elif isinstance(data, int):
        assert isinstance(real, int)
        assert data == real
    elif isinstance(data, float):
        assert isinstance(real, float)
        data, p1 = str2float(data)
        real, p2 = str2float(real)
        p = min(p1, p2)
        assert round(data, p) == round(real, p)
    else:
        assert data == real


def timestamp():
    # js 格式的时间戳
    return int(time.time() * 1000)



from copy import deepcopy


def rule(data):
    result = []
    for k in data:
        result.append(data[k]['code'])
    return result


def optimum(sub, result, path):
    if path != '/':
        path += '.'

    res = result
    for k in sub:
        result = deepcopy(res)
        temp = res
        res = []
        flag = False
        for data in result:
            if path + k not in data:
                flag = True
                res.append(data)
        if not flag:
            temp.sort(key=lambda d: rule(d))
            return temp[0]
    result.sort(key=lambda d: rule(d))
    return result[0]


def list2dict(data):
    keys = ['['+str(i)+']' for i in range(len(data))]
    return dict(zip(keys, data))


def check(sub, parent,  sp='/', pp='/'):
    '''
    sp: sub_path
    pp: parent_path
    '''
    re = {'code': 0, 'result': {}, 'var': {}, 'none':[]}
    if sp != '/':
        sp += '.'
    if pp != '/':
        pp += '.'
  
    def _in(k, data):
        try:
            return  eval('data[k]')
        except:
            return ''      


    for k, sv in sub.items():
        # 判断键值是否是 <value> 格式，如果是，则表明是变量赋值
        var_flag = isinstance(sv, str) and sv.startswith(
            '<') and sv.endswith('>')
        index = ''
        _k = k

        if '.' in k:
            s = k.split('.')
            for _s in s[1:]:
                if '[' in _s:
                    d = _s.split('[', 1)
                    index += '[\'' + d[0] + '\']' + '[\''  + d[1]
                else:
                    index += '[\'' + _s + '\']'
            if '[' in s[0]:
                k = s[0].split('[', 1)[0]
                index = '[' + s[0].split('[', 1)[1] + index
            else:
                k = s[0]

        elif '[' in k:
            s = k.split('[', 1)
            index = '[' + s[1]
            k = s[0]

        # 预期键不存在
        if sv == '-':
            # 预期键不存在，实际键存在
            if k in parent:
                re['result'][sp + _k] = {'code': 4, 'sv': sv, 'pp': pp + _k, 'pv': parent[k]}

        # 预期键存在
        elif sv == '+':
            # 预期键存在，实际键不存在
            if k not in parent:
                re['result'][sp + _k] = {'code': 3, 'sv': sv, 'pp': None, 'pv': None}

        elif k in parent:
            if index:
                try:
                    pv = eval('parent[k]' + index)
                except:
                    if var_flag:
                        re['var'][sv[1:-1]] = None
                        re['none'].append(sv[1:-1])
                    else:                    
                        re['result'][sp + _k] = {'code': 3, 'sv': sv, 'pp': None, 'pv': None}
                    continue
            else:   
                pv = parent[k]

            code = 0

            if var_flag:
                re['var'][sv[1:-1]] = pv
                continue

            elif isinstance(sv, str):
                if sv.startswith('#'):
                    if sv[1:] == str(pv):
                        code = 1
                elif sv.startswith('<>'):
                    if (isinstance(pv, int) or isinstance(pv, float)) and pv == float(sv[2:]):
                        code = 1                          
                elif sv.startswith('>='):
                    if (isinstance(pv, int) or isinstance(pv, float)) and pv < float(sv[2:]):
                        code = 1                        
                elif sv.startswith('>'):
                    if (isinstance(pv, int) or isinstance(pv, float)) and pv <= float(sv[1:]):
                        code = 1
                elif sv.startswith('<='):
                    if (isinstance(pv, int) or isinstance(pv, float)) and pv > float(sv[2:]):
                        code = 1                        
                elif sv.startswith('<'):
                    if (isinstance(pv, int) or isinstance(pv, float)) and pv >= float(sv[1:]):
                        code = 1                                                                         
                elif not isinstance(pv, str):
                    code = 2  # 键值的数据类型不一致
                elif sv.startswith('*'):
                    if sv[1:] not in pv:
                        code = 1
                elif sv.startswith('^'):
                    if not pv.startswith(sv[1:]):
                        code = 1
                elif sv.startswith('$'):
                    if not pv.endswith(sv[1:]):
                        code = 1
                elif sv.startswith('\\'):
                    sv = sv[1:]
                elif sv != pv:
                    code = 1  # 键值不等

            elif isinstance(sv, int):
                if not isinstance(pv, int):
                    code = 2  # 键值的数据类型不一致
                elif sv != pv:
                    code = 1  # 键值不等

            elif isinstance(sv, float):
                if not isinstance(pv, float):
                    code = 2  # 键值的数据类型不一致
                elif sv != pv:
                    code = 1  # 键值不等

            elif isinstance(sv, list):
                if not isinstance(pv, list):
                    code = 2  # 键值的数据类型不一致
            
                else:
                    for i in range(len(sv)):  # 把二级列表转换为 dict                         
                        if isinstance(sv[i], list):
                            sv[i] = list2dict(sv[i])
                    for i in range(len(pv)):  # 把二级列表转换为 dict                            
                        if isinstance(pv[i], list):
                            pv[i] = list2dict(pv[i]) 

                    if isinstance(sv[0], dict):  # list 子项为 dict
                        for i, sv_i in enumerate(sv):
                            result = []
                            flag = False
                            for j, pv_i in enumerate(pv):
                                r = check(sv_i, pv_i, sp + _k + '[%s]' % i, pp + _k + '[%s]' % j)
                                if r['code'] == 0:
                                    flag = True
                                    re['var'] = dict(re['var'], **r['var'])
                                    break
                                else:
                                    result.append(r['result'])
                            if result:
                                o = optimum(sv_i, result, sp + k + '[%s]' % i)
                            else:
                                o = {}
                            re['var'] = dict(re['var'], **re['var'])

                            if not flag:
                                re['result'] = dict(re['result'], **o)

                    else:  # list 子项为 int/str/float/None/False/True
                        for v in sv:
                            if v not in pv:
                                code = 5  # 预期的 list 值在实际值的 list 不存在
                                re['result'][sp + _k] = {'code': 5, 'sv': sv, 'pp': pp + _k, 'pv': pv}

            elif isinstance(sv, dict):
                if not isinstance(pv, dict):
                    code = 2  # 键值的数据类型不一致
                else:
                    r = check(sv, pv, sp + k, pp + k)
                    if r['code'] == 0:
                       re['var'] = dict(re['var'], **r['var'])
                       continue
                    else:
                        re['result'] = dict(re['result'], **r['result'])
                        for k in r['var']:
                            r['var'][k] = None
                            if k not in re['none']:
                                re['none'].append(k)
                        re['var'] = dict(re['var'], **r['var'])


            if code != 0:
                re['result'][sp + _k] = {'code': code, 'sv': sv, 'pp': pp + _k, 'pv': pv}
        else:  # 键不存在
            if var_flag:
                re['var'][sv[1:-1]] = None
                re['none'].append(sv[1:-1])
            else:            
                re['result'][sp + _k] = {'code': 3, 'sv': sv, 'pp': None, 'pv': None}

    re['code'] = len(re['result'])
    return re




import logging
import datetime
from pathlib import Path
import sys
from qingtest.utility import mkdir


def today():
    now = datetime.datetime.now()
    return now.strftime('%Y%m%d')



def set_log(logger, log_path):
    mkdir('log')

    # 文件日志
    log_file = Path('log') / f'{today()}.log'
    file_handler = logging.FileHandler(filename=log_file, encoding="utf-8")
    file_handler.setFormatter(formatter)  # 可以通过setFormatter指定输出格式

    # 单次文件日志
    sweet_log = log_path / 'sweet.log'   
    try:
        sweet_log.unlink()
    except:
        pass
    sweet_handler = logging.FileHandler(filename=sweet_log, encoding="utf-8")
    sweet_handler.setFormatter(formatter)  # 可以通过setFormatter指定输出格式

    # 控制台日志
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.formatter = formatter  # 也可以直接给formatter赋值

    # 为logger添加的日志处理器
    logger.addHandler(file_handler)
    logger.addHandler(sweet_handler)    
    logger.addHandler(console_handler)

    # 指定日志的最低输出级别，默认为WARN级别
    # DEBUG，INFO，WARNING，ERROR，CRITICAL
    logger.setLevel(logging.INFO)

    return str(sweet_log)

# 获取logger实例，如果参数为空则返回root logger
logger = logging.getLogger("qingtest")
# 指定logger输出格式
formatter = logging.Formatter('%(asctime)s [%(levelname)s]: #  %(message)s')


import time
from selenium import webdriver
from qingtest.config import element_wait_timeout, page_flash_timeout


def now():
    t = time.time()
    return time.strftime("@%Y%m%d_%H%M%S", time.localtime(t))


def timestamp():
    # js 格式的时间戳
    return int(time.time() * 1000)


class Global:
    def __init__(self):
        self.start_time = now()
        self.start_timestamp = timestamp()
        self.plan_name = ''
        self.sheet_name = ''
        self.plan_data = {}
        self.testsuite_data = {}
        self.no = 1
        self.driver = ''
        self.snippet = {}
        self.caseset = {}

    def init(self, desired_caps, server_url):
        self.desired_caps = desired_caps
        self.server_url = server_url
        self.platform = desired_caps.get('platformName', '')
        self.browserName = desired_caps.get('browserName', '')
        self.headless = desired_caps.pop('headless', False)
        self.snapshot = desired_caps.pop('snapshot', False)
        self.executable_path = desired_caps.pop('executable_path', False)
        self.mobile = desired_caps.pop('mobile', False)

    def set_driver(self):
        self.test_data = {'_last_': False}
        self.var = {}
        self.casesets = []  # 用例组合执行容器
        self.current_page = '通用'
        self.db = {}
        self.http = {}
        self.windows = {}
        self.baseurl = {}
        self.action = {}
        self.wait_times = 0

        if self.platform.lower() == 'desktop':
            if self.browserName.lower() == 'ie':
                #capabilities = webdriver.DesiredCapabilities().INTERNETEXPLORER
                #capabilities['acceptInsecureCerts'] = True
                if self.executable_path:
                    self.driver = webdriver.Ie(
                        executable_path=self.executable_path)
                else:
                    self.driver = webdriver.Ie()
            elif self.browserName.lower() == 'firefox':
                profile = webdriver.FirefoxProfile()
                profile.accept_untrusted_certs = True

                options = webdriver.FirefoxOptions()
                # 如果配置了 headless 模式
                if self.headless:
                    options.set_headless()
                    # options.add_argument('-headless')
                    options.add_argument('--disable-gpu')
                    options.add_argument("--no-sandbox")
                    options.add_argument('window-size=1920x1080')

                if self.executable_path:
                    self.driver = webdriver.Firefox(
                        firefox_profile=profile, firefox_options=options, executable_path=self.executable_path)
                else:
                    self.driver = webdriver.Firefox(
                        firefox_profile=profile, firefox_options=options)
                self.driver.maximize_window()
            elif self.browserName.lower() == 'chrome':
                options = webdriver.ChromeOptions()
                if self.mobile == True:
                    if self.headless:
                        options.add_argument('--headless')
                        options.add_argument('--disable-gpu')
                        options.add_argument("--no-sandbox")
                        options.add_argument('window-size=375,812')  # 模拟手机分辨率
                        options.add_argument('--user-agent=iphone x')  # 模拟手机ua
                    options.add_argument("--start-maximized")
                    options.add_argument('--ignore-certificate-errors')
                    # 指定浏览器分辨率，当"--start-maximized"无效时使用
                    options.add_argument('window-size=375,812')# 模拟手机分辨率
                    options.add_argument('--user-agent=iphone x')# 模拟手机ua
                    prefs = {}
                    prefs["credentials_enable_service"] = False
                    prefs["profile.password_manager_enabled"] = False
                    options.add_experimental_option("prefs", prefs)
                    options.add_argument('disable-infobars')
                    options.add_experimental_option(
                        "excludeSwitches", ['load-extension', 'enable-automation'])
                if self.mobile == False:
                    if self.headless:
                        options.add_argument('--headless')
                        options.add_argument('--disable-gpu')
                        options.add_argument("--no-sandbox")
                        options.add_argument('window-size=1920x1080')

                    options.add_argument("--start-maximized")
                    options.add_argument('--ignore-certificate-errors')
                    # 指定浏览器分辨率，当"--start-maximized"无效时使用
                    prefs = {}
                    prefs["credentials_enable_service"] = False
                    prefs["profile.password_manager_enabled"] = False
                    options.add_experimental_option("prefs", prefs)
                    options.add_argument('disable-infobars')
                    options.add_experimental_option(
                    "excludeSwitches", ['load-extension', 'enable-automation'])
                if self.executable_path:
                    self.driver = webdriver.Chrome(
                        options=options, executable_path=self.executable_path)
                else:
                    self.driver = webdriver.Chrome(options=options)

            else:
                raise Exception(
                    'Error: this browser is not supported or mistake name：%s' % self.browserName)
            # 等待元素超时时间
            self.driver.implicitly_wait(element_wait_timeout)  # seconds
            # 页面刷新超时时间
            self.driver.set_page_load_timeout(page_flash_timeout)  # seconds

        elif self.platform.lower() == 'ios':
            from appium import webdriver as appdriver
            if not self.driver:
                self.driver = appdriver.Remote(
                    self.server_url, self.desired_caps)

        elif self.platform.lower() == 'android':
            from appium import webdriver as appdriver
            if not self.driver:
                self.driver = appdriver.Remote(
                    self.server_url, self.desired_caps)

        elif self.platform.lower() == 'windows':
            from pywinauto.application import Application
            from qingtest.keywords.windows import Windows
            self.desired_caps.pop('platformName')
            backend = self.desired_caps.pop('backend', 'win32')
            _path = ''
            if self.desired_caps.get('#path'):
                _path = self.desired_caps.pop('#path')
                _backend = self.desired_caps.pop('#backend')

            if self.desired_caps.get('cmd_line'):
                app = Application(backend).start(**self.desired_caps)
            elif self.desired_caps.get('path'):
                app = Application(backend).connect(**self.desired_caps)
            else:
                raise Exception('Error: Windows GUI start/connect args error')
            self.windows['default'] = Windows(app)

            if _path:
                _app = Application(_backend).connect(path=_path)
                self.windows['#'] = Windows(_app)

    def plan_end(self):
        self.plan_data['plan'] = self.plan_name
        #self.plan_data['task'] = self.start_timestamp
        self.plan_data['start_timestamp'] = self.start_timestamp
        self.plan_data['end_timestamp'] = int(time.time() * 1000)

        return self.plan_data


g = Global()



from qingtest.utility import Excel, data2dict, replace
from qingtest.log import logger


def elements_format(data):
    elements = {}
    page = ''
    custom = ''
    for d in data:
        if d['page'].strip():
            page = d['page']
            custom = ''
        else:
            d['page'] = page

        if d.get('custom', '').strip():
            custom = d['custom']
        else:
            d['custom'] = custom

        elements[d['page'] + '-' + d['element']] = d
    return elements


class Elements:
    def __init__(self):
        pass

    def env(self):
        pass

    def get_elements(self, elements_file):
        d = Excel(elements_file)
        self.elements = elements_format(data2dict(d.read('elements')))

    def have(self, page, element):
        ele = element.split('#')

        if len(ele) >= 2:
            _el = ele[0] + '#'
        else:
            _el = element
        # 如果有<>,则不判断了
        if '<' in _el:
            return '', '通用' + '-' + element
        # 在元素定位表中查询
        elem = page + '-' + _el
        if self.elements.get(elem, ''):
            return self.elements[elem]['custom'], page + '-' + element
        else:
            # 查不到就在通用里查,还是查不到，可能是不在 element.xlsx 中定义的元素
            elem = '通用' + '-' + _el
            if self.elements.get(elem, ''):
                return self.elements[elem]['custom'], '通用' + '-' + element
            else:
                logger.info('Page:%s element:%s' % (page, element))
                return '', element

    def get(self, element, flag=False):
        ele = element.split('#')
        # #号后面的值，即用户输入的变量
        _v = []
        # 支持多个变量替代，但是顺序要对应
        if len(ele) >= 2:
            _el = ele[0] + '#'
            _v = ele[1:]
        else:
            _el = element
        el = self.elements.get(_el, '')
        if not el:
            if flag:
                return _el, ''
            return _el, element.split('#', 1)[-1]
        value = el['value']
        for v in _v:
            v = '#' if v=='^' else v  # 当 value 中的 # 无需替换时，用例中的元素使用 ^ 表示
            value = value.replace('#', v, 1)
        return el, replace(value)


e = Elements()



import sys
import shutil
import zipfile
from pathlib import Path
from qingtest.autotest import Autotest
from qingtest.report import reporter


def extract(zfile, path):
    f = zipfile.ZipFile(zfile, 'r')
    for file in f.namelist():
        f.extract(file, path)


def qingtest():
    qingtest_dir = Path(__file__).resolve().parents[0]
    example_dir = qingtest_dir / 'example' / 'qingtest_example.zip'
    extract(str(example_dir), Path.cwd())

    print('\n文档: https://sweeter.io\n公众号：喜文测试\nQ Q 群：941761748 (验证码：python)注意首字母小写')
    print('\n\n生成 qingtest example 成功\n快速体验，请输入如下命令(进入示例目录，启动运行脚本):\n\ncd qingtest_example\npython start.py')


def report():
    qingtest_dir = Path(__file__).resolve().parents[0]
    report_dir = qingtest_dir / 'example' / 'report.zip'
    extract(str(report_dir), Path.cwd())


from copy import deepcopy
import requests
import json
from qingtest.injson import check
from qingtest.globals import g
from qingtest.elements import e
from qingtest.log import logger
from qingtest.parse import data_format
from qingtest.utility import json2dict
from pathlib import Path

path = Path('lib') / 'http_handle.py'
if path.is_file():
    from lib import http_handle
else:
    from qingtest.lib import http_handle


class Http:

    def __init__(self, step):
        # 获取 baseurl
        baseurl = e.get(step['page'] + '-' + 'baseurl', True)[1]
        if not baseurl:
            self.baseurl = ''
        else:
            if not baseurl.endswith('/'):
                baseurl += '/'
            self.baseurl = baseurl

        self.r = requests.Session()
        # 获取 headers
        self.headers_get = e.get(step['page'] + '-' + 'headers_get', True)[1]
        self.headers_post = e.get(step['page'] + '-' + 'headers_post', True)[1]


def get(step):
    request('get', step)


def post(step):
    request('post', step)


def put(step):
    request('put', step)


def patch(step):
    request('patch', step)


def delete(step):
    request('delete', step)


def options(step):
    request('options', step)


def request(kw, step):
    element = step['element']
    url = e.get(element)[1]
    if url.startswith('/'):
        url = url[1:]

    data = step['data']
    # 测试数据解析时，会默认添加一个 text 键，需要删除
    if 'text' in data and not data['text']:
        data.pop('text')

    _data = {}
    _data['headers'] = json2dict(data.pop('headers', '{}'))
    if data.get('cookies'):
        data['cookies'] = json2dict(data['cookies'])
    if kw == 'get':
        _data['params'] = json2dict(
            data.pop('params', '{}')) or json2dict(data.pop('data', '{}'))
    elif kw == 'post':
        if data.get('text'):
            _data['data'] = data.pop('text').encode('utf-8')
        else:
            _data['data'] = json2dict(data.pop('data', '{}'))
        _data['json'] = json2dict(data.pop('json', '{}'))
        _data['files'] = eval(data.pop('files', 'None'))
    elif kw in ('put', 'patch'):
        _data['data'] = json2dict(data.pop('data', '{}'))


    for k in data:
        for s in ('{', '[', 'False', 'True'):
            if s in data[k]:
                try:
                    data[k] = eval(data[k])
                except:
                    logger.warning('Try eval data failure: %s' % data[k])
                break
    expected = step['expected']
    expected['status_code'] = expected.get('status_code', None)
    expected['text'] = expected.get('text', None)
    expected['json'] = json2dict(expected.get('json', '{}'))
    expected['cookies'] = json2dict(expected.get('cookies', '{}'))
    expected['headers'] = json2dict(expected.get('headers', '{}'))
    timeout = float(expected.get('timeout', 10))
    expected['time'] = float(expected.get('time', 0))

    if not g.http.get(step['page']):
        g.http[step['page']] = Http(step)
    http = g.http[step['page']]

    if kw == 'post':
        if http.headers_post:
            http.r.headers.update(eval(http.headers_post))
    else:
        if http.headers_get:
            http.r.headers.update(eval(http.headers_get))

    logger.info('URL: %s' % http.baseurl + url)

    # 处理 before_send
    before_send = data.pop('before_send', '')
    if before_send:
        _data, data = getattr(http_handle, before_send)(kw, _data, data)
    else:
        _data, data = getattr(http_handle, 'before_send')(kw, _data, data)

    if _data['headers']:
        for k in [x for x in _data['headers']]:
            if not _data['headers'][k]:
                del http.r.headers[k]
                del _data['headers'][k]
        http.r.headers.update(_data['headers'])

    if kw == 'get':
        r = getattr(http.r, kw)(http.baseurl + url,
                                params=_data['params'], timeout=timeout, **data)
        if _data['params']:
            logger.info(f'PARAMS: {_data["params"]}')

    elif kw == 'post':
        r = getattr(http.r, kw)(http.baseurl + url,
                                data=_data['data'], json=_data['json'], files=_data['files'], timeout=timeout, **data)
        logger.info(f'BODY: {r.request.body}')

    elif kw in ('put', 'patch'):
        r = getattr(http.r, kw)(http.baseurl + url,
                                data=_data['data'], timeout=timeout, **data)
        logger.info(f'BODY: {r.request.body}')
                                
    elif kw in ('delete', 'options'):
        r = getattr(http.r, kw)(http.baseurl + url, timeout=timeout, **data)

    logger.info('status_code: %s' % repr(r.status_code))
    try:  # json 响应
        logger.info('response json: %s' % repr(r.json()))
    except:  # 其他响应
        logger.info('response text: %s' % repr(r.text))

    response = {'status_code': r.status_code, 'headers': r.headers,
                '_cookies': r.cookies, 'content': r.content, 'text': r.text}

    try:
        response['cookies'] = requests.utils.dict_from_cookiejar(r.cookies)
    except:
        response['cookies'] = r.cookies

    try:
        j = r.json()
        response['json'] = j
    except:
        response['json'] = {}

    # 处理 after_receive
    after_receive = expected.pop('after_receive', '')
    if after_receive:
        response = getattr(http_handle, after_receive)(response)
    else:
        response = getattr(http_handle, 'after_receive')(response)

    var = {}  # 存储所有输出变量

    if expected['status_code']:
        if str(expected['status_code']) != str(response['status_code']):
            raise Exception(f'status_code | EXPECTED:{repr(expected["status_code"])}, REAL:{repr(response["status_code"])}')

    if expected['text']:
        if expected['text'].startswith('*'):
            if expected['text'][1:] not in response['text']:
                raise Exception(f'text | EXPECTED:{repr(expected["text"])}, REAL:{repr(response["text"])}')
        else:
            if expected['text'] == response['text']:
                raise Exception(f'text | EXPECTED:{repr(expected["text"])}, REAL:{repr(response["text"])}')

    if expected['headers']:
        result = check(expected['headers'], response['headers'])
        logger.info('headers check result: %s' % result)
        if result['code'] != 0:
            raise Exception(f'headers | EXPECTED:{repr(expected["headers"])}, REAL:{repr(response["headers"])}, RESULT: {result}')
        elif result['var']:
            var = dict(var, **result['var'])
            g.var = dict(g.var, **result['var'])
            logger.info('headers var: %s' % (repr(result['var']))) 

    if expected['cookies']:
        logger.info('response cookies: %s' % response['cookies'])
        result = check(expected['cookies'], response['cookies'])
        logger.info('cookies check result: %s' % result)
        if result['code'] != 0:
            raise Exception(f'cookies | EXPECTED:{repr(expected["cookies"])}, REAL:{repr(response["cookies"])}, RESULT: {result}')
        elif result['var']:
            var = dict(var, **result['var'])
            g.var = dict(g.var, **result['var'])
            logger.info('cookies var: %s' % (repr(result['var']))) 

    if expected['json']:
        result = check(expected['json'], response['json'])
        logger.info('json check result: %s' % result)
        if result['code'] != 0:
            raise Exception(f'json | EXPECTED:{repr(expected["json"])}, REAL:{repr(response["json"])}, RESULT: {result}')
        elif result['var']:
            var = dict(var, **result['var'])
            g.var = dict(g.var, **result['var'])
            logger.info('json var: %s' % (repr(result['var'])))            

    if expected['time']:
        if expected['time'] < r.elapsed.total_seconds():
            raise Exception(f'time | EXPECTED:{repr(expected["time"])}, REAL:{repr(r.elapsed.total_seconds())}')

    output = step['output']
    # if output:
    #     logger.info('output: %s' % repr(output))

    for k, v in output.items():
        if v == 'status_code':
            g.var[k] = response['status_code']
            logger.info('%s: %s' % (k, repr(g.var[k])))
        elif v == 'text':
            g.var[k] = response['text']
            logger.info('%s: %s' % (k, repr(g.var[k])))
        elif k == 'json':
            sub = json2dict(output.get('json', '{}'))
            result = check(sub, response['json'])
            # logger.info('Compare json result: %s' % result)
            var = dict(var, **result['var'])
            g.var = dict(g.var, **result['var'])
            logger.info('json var: %s' % (repr(result['var'])))
        elif k == 'cookies':
            sub = json2dict(output.get('cookies', '{}'))
            result = check(sub, response['cookies'])
            # logger.info('Compare json result: %s' % result)
            var = dict(var, **result['var'])
            g.var = dict(g.var, **result['var'])
            logger.info('cookies var: %s' % (repr(result['var'])))
    if var:
        step['_output'] += '\n||output=' + str(var)


from qingtest.log import logger


class DB:

    def __init__(self, arg):
        self.connect = ''
        self.cursor = ''
        self.db = ''

        try:
            if arg['type'].lower() == 'mongodb':
                import pymongo
                host = arg.pop('host') if arg.get('host') else 'localhost:27017'
                host = host.split(',') if ',' in host else host
                port = int(arg.pop('port')) if arg.get('port') else 27017
                if arg.get('user'):
                    arg['username'] = arg.pop('user')
                # username = arg['user'] if arg.get('user') else ''
                # password = arg['password'] if arg.get('password') else ''
                # self.connect = pymongo.MongoClient('mongodb://' + username + password + arg['host'] + ':' + arg['port'] + '/')
                self.connect = pymongo.MongoClient(host=host, port=port, **arg)
                self.connect.server_info()
                self.db = self.connect[arg['dbname']]
                
                return

            if arg['type'].lower() == 'mysql':
                import pymysql as mysql
                self.connect = mysql.connect(
                    host=arg['host'], port=int(arg['port']), user=arg['user'], password=arg['password'], database=arg['dbname'], charset=arg.get('charset', 'utf8'))
                self.cursor = self.connect.cursor()
                sql = 'select version()'

            elif arg['type'].lower() == 'oracle':
                import os
                import cx_Oracle as oracle
                # Oracle查询出的数据，中文输出问题解决
                os.environ['NLS_LANG'] = 'SIMPLIFIED CHINESE_CHINA.UTF8'
                self.connect = oracle.connect(
                    arg['user'] + '/' + arg['password'] + '@' + arg['host'] + '/' + arg['sid'])
                self.cursor = self.connect.cursor()
                sql = 'select * from v$version'
            elif arg['type'].lower() == 'sqlserver':
                import pymssql as sqlserver
                self.connect = sqlserver.connect(
                    host=arg['host'], port=arg['port'], user=arg['user'], password=arg['password'], database=arg['dbname'], charset=arg.get('charset', 'utf8'))
                self.cursor = self.connect.cursor()
                sql = 'select @@version'

            self.cursor.execute(sql)
            self.cursor.fetchone()

        except:
            logger.exception('*** %s connect is failure ***' % arg['type'])
            raise

    def fetchone(self, sql):
        try:
            self.cursor.execute(sql)
            data = self.cursor.fetchone()
            self.connect.commit()
            return data
        except:
            logger.exception('*** Fetchone failure ***')
            raise

    def fetchall(self, sql):
        try:
            self.cursor.execute(sql)
            data = self.cursor.fetchall()
            self.connect.commit()
            return data
        except:
            logger.exception('*** Fetchall failure ***')
            raise

    def execute(self, sql):
        try:
            self.cursor.execute(sql)
            self.connect.commit()
        except:
            logger.exception('*** Execute failure ***')
            raise


    def mongo(self, collection, sql):
        try:
            cmd = 'self.db[\'' + collection + '\'].' + sql
            result = eval(cmd)
            if sql.startswith('find_one'):
                return result
            elif sql.startswith('find'):
                for d in result:
                    return d
            elif 'count' in sql:
                return {'count': result}
            else:
                return {}
        except:
            logger.exception('*** Execute failure ***')
            raise            

    def __del__(self):
        self.connect.close()



from time import sleep
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from qingtest.elements import e
from qingtest.globals import g
from qingtest.windows import w
from qingtest.log import logger
from qingtest.config import element_wait_timeout


def locating_element(element, action=''):
    el_location = None
    try:
        el, value = e.get(element)
    except:
        logger.exception(
            'Locating the element:%s is Failure, this element is not define' % element)
        raise Exception(
            'Locating the element:%s is Failure, this element is not define' % element)

    if not isinstance(el, dict):
        raise Exception(
            'Locating the element:%s is Failure, this element is not define' % element)

    wait = WebDriverWait(g.driver, element_wait_timeout)

    if el['by'].lower() in ('title', 'url', 'current_url'):
        return None
    else:
        try:
            el_location = wait.until(EC.presence_of_element_located(
                (getattr(By, el['by'].upper()), value)))
        except:
            sleep(5)
            try:
                el_location = wait.until(EC.presence_of_element_located(
                    (getattr(By, el['by'].upper()), value)))            
            except :
                raise Exception('Locating the element:%s is Failure: Timeout' % element)                
    try:
        if g.driver.name in ('chrome', 'safari'):
            g.driver.execute_script(
                "arguments[0].scrollIntoViewIfNeeded(true)", el_location)
        else:
            g.driver.execute_script(
                "arguments[0].scrollIntoView(false)", el_location)
    except:
        pass

    try:
        if action == 'CLICK':
            el_location = wait.until(EC.element_to_be_clickable(
                (getattr(By, el['by'].upper()), value)))
        else:
            el_location = wait.until(EC.visibility_of_element_located(
                (getattr(By, el['by'].upper()), value)))
    except:
        pass

    return el_location


def locating_elements(elements):
    elements_location = {}
    for el in elements:
        elements_location[el] = locating_element(el)
    return elements_location


def locating_data(keys):
    data_location = {}
    for key in keys:
        data_location[key] = locating_element(key)
    return data_location



import time
import arrow
from pathlib import Path
from qingtest.globals import g
from qingtest.utility import mkdir


def reporter(plan_data, testsuites_data, report_data, extra_data):

    extra_data['plan'] = plan_data['plan']
    extra_data['task'] = int(time.time() * 1000)

    testcases = []
    for key, ts in report_data.items():
        count = {'result': 'success', 'total': 0,
                 'success': 0, 'failure': 0, 'blocked': 0}
        no = 1
        for tc in ts:
            tc['testsuite'] = key
            tc['no'] = no
            no += 1
            tc = {**extra_data, **tc}

            res = tc['result'].lower()
            if tc['condition'].lower() in ('base', 'setup', 'snippet'):
                pass
            elif res in count:
                count[res] += 1
                count['total'] += 1
            testcases.append(tc)
            if count['failure'] + count['blocked']:
                count['result'] = 'failure'
        testsuites_data[key] = {**count, **testsuites_data[key]}

    testsuite = []
    count = {'total': 0, 'success': 0,
             'failure': 0, 'blocked': 0}
    result = 'success'
    for key, ts in testsuites_data.items():
        for k in count:
            count[k] += ts[k]
        if ts['result'] != 'success':
            result = 'failure'

        ts = {**extra_data, **ts}
        ts['testsuite'] = key
        testsuite.append(ts)
    count['result'] = result

    plan = {**extra_data, **count, **plan_data}

    return plan, testsuite, testcases


def local_time(timestamp):
    import time
    t = time.localtime(int(timestamp / 1000))
    return str(time.strftime("%Y/%m/%d %H:%M:%S", t))


def cost_time(start, end):
    return int((end - start) / 1000)


def summary(plan_data, testsuites_data, report_data, extra_data):
    plan, testsuites, testcases = reporter(
        plan_data, testsuites_data, report_data, extra_data)

    data = [['测试套件', '用例总数', '成功', '阻塞', '失败', '测试结果', '开始时间', '结束时间', '耗时(秒)']]
    failures = [['测试套件', '用例编号', '用例标题', '用例结果', '失败步骤', '备注']]

    for suite in testsuites:
        row = [suite['testsuite'], suite['total'], suite['success'], suite['blocked'], suite['failure'],
               suite['result'], local_time(suite['start_timestamp']), local_time(suite['end_timestamp']),
               cost_time(suite['start_timestamp'], suite['end_timestamp'])]
        data.append(row)

    flag = False
    for case in testcases:
        suite_name = '' if flag else case['testsuite']
        row = []
        if case['result'] == 'blocked':
            row = [suite_name, case['id'], case['title'], case['result']]
        elif case['result'] == 'failure':
            for step in case['steps']:
                if step['score'] == 'NO':
                    desc = '|'.join([step[k] for k in ('no', 'keyword', 'page', 'element')])
                    row = [suite_name, case['id'], case['title'], case['result'], desc, step['remark']]
                    break
        if row:
            flag = True
            failures.append(row)

    data.append(['--------'])
    total = ['总计', plan['total'], plan['success'], plan['blocked'], plan['failure'],
             plan['result'], local_time(plan['start_timestamp']), local_time(plan['end_timestamp']),
             cost_time(plan['start_timestamp'], plan['end_timestamp'])]

    data.append(total)
    if len(failures) > 1:
        data.append(['********'])
        data += failures
    return data

def markdown(plan, testsuites, testcases, md_path='markdown'):
    success = OK = '<font color=#00BB00>通过</font>'
    failure = NO = '<font color=#FF0000>失败</font>'
    blocked = '<font color=#FFD306>阻塞</font>'
    skipped = '<font color=#6C6C6C>-</font>'

    md = '| 测试套件名称 | 开始时间 | 结束时间 | 耗时 | 成功个数 | 失败个数 | 阻塞个数 | 总个数 | 结果 |\n'
    md +='| ----------- | ------- | ------- | ---- | ------- | ------- | -------- | ----- | ---- |\n'

    result = success
    sc, fc, bc, tc = 0, 0, 0, 0
    for v in testsuites.values():
        sc += v['success']
        fc += v['failure'] 
        bc += v['blocked']
        tc += v['total']
        re = success
        if v['result'] == 'failure':
            re = failure
        cost = round((v['end_timestamp'] - v['start_timestamp'])/1000, 1)  
        md += f'| {v["testsuite"]} | {tm(v["start_timestamp"])} | {tm(v["end_timestamp"])} | {cost} | '
        md += f'{v["success"]} | {v["failure"]} | {v["blocked"]} | {v["total"]} | {re} |\n'
        if v['result'] == 'failure':
            result = failure
    cost = round((plan['end_timestamp'] - plan['start_timestamp'])/1000, 1)
    md += f'| **共计** | {tm(plan["start_timestamp"])} | {tm(plan["end_timestamp"])}  | {cost} | '
    md += f'{sc} | {fc} | {bc} | {tc} | {result} |\n'
    title = f'# 「{plan["plan"]}」自动化测试执行报告 {result} #\n\n[历史记录](/{plan["plan"]}/)\n\n'
    md = title + f'## 测试计划执行结果\n\n{md}\n\n## 测试套件执行结果\n\n'

    if result == success:
        icon = '✔️'
    else:
        icon = '❌'

    message = f'- {icon} <font color=#9D9D9D size=2>{tm(plan["start_timestamp"])} - {tm(plan["end_timestamp"])}</font> 测试计划'
    message +=f'「[{plan["plan"]}]({plan["plan"]}/{plan["plan"]}_{tm(plan["start_timestamp"], "_")})」执行完成，测试结果：{result}，成功：{sc}，失败：{fc}，阻塞：{bc}\n\n'

    # 测试套件 - 测试用例结果
    txt = ''
    for k,v in testcases.items():
        txt += f'\n- ### {k}\n\n'
        txt += '| 用例id  | 用例名称 |   前置条件   |开始时间         | 结束时间       | 耗时   | 结果    |\n'
        txt += '| ------- | ------- | ----------- | -------------- | -------------- | ----- | ------- |\n'         
        for case in v:
            if case['flag'] == 'N':
                continue
            cost = round((case['end_timestamp'] - case['start_timestamp'])/1000, 1)
            result = eval(case['result'])
            txt += f'| [{case["id"]}](#{case["id"]}) | {case["title"]} | {case["condition"]} | {tm(case["start_timestamp"])} | {tm(case["end_timestamp"])} | {cost} | {result} |\n'
            
    md += f'{txt}\n\n## 测试用例执行结果\n'
    txt = ''
    for k,v in testcases.items():    
        txt += f'\n- ### {k}\n'
        for case in v:
            if case['flag'] == 'N':
                continue               
            txt += f'\n#### {case["id"]}\n\n**{case["title"]}** | {case["condition"]} | {case["designer"]} | {eval(case["result"])}\n\n'
            txt += '| 步骤  | 操作  | 页面  | 元素  | 测试数据  | 预期结果 | 输出数据  | 耗时 | 测试结果 | 备注 | 截图   |\n'
            txt += '|------|-------|-------|------|-----------|---------|-----------|-----|---------|------|--------|\n'
            for step in case['steps']:
                cost = round((step.get('end_timestamp', 0) - step.get('start_timestamp', 0))/1000, 1)
                if cost == 0:
                    cost = '-'
                if not step['score']:
                    result = skipped
                else:
                    result = eval(step['score'])
                snapshot = ''
                if 'snapshot' in step:
                    for k,v in step['snapshot'].items():
                        snapshot += f"[{k}](/report/{v} ':ignore')\n"
                txt += f'| {step["no"]} | {step["keyword"]} | {step["page"]} | {escape(step["element"])} | {escape(step["data"])} | {escape(step["expected"])} | {escape(step["output"])} | {cost} | {result} | {step["remark"]} | {escape(snapshot, "%23")} |\n'
            result = eval(case['result'])            
    md += txt            


    p = Path(md_path) / 'report'
    latest = p / 'latest'
    report = p  / g.plan_name
    mkdir(p)
    mkdir(report)


    with open(p / 'README.md', 'r', encoding='UTF-8') as f:
        txt = f.read()
        if '恭喜你安装成功' in txt:
            txt = ''
    with open(p / f'README.md', 'w', encoding='UTF-8') as f:      
        f.write(message + txt)
    with open(latest / f'{g.plan_name}.md','w', encoding='UTF-8') as f:
        f.write(md)
    readme = report / 'README.md'
    if readme.is_file():
        with open(report / 'README.md', 'r', encoding='UTF-8') as f:
            txt = f.read()
    else:
        txt = ''        
    with open(report / 'README.md','w', encoding='UTF-8') as f:
        f.write(message + txt)        
    with open(report / f'{plan["plan"]}_{tm(plan["start_timestamp"], "_")}.md','w',encoding='UTF-8') as f:
        f.write(md)
    with open(p / '_sidebar.md', 'r', encoding='UTF-8') as f:
        txt = f.read()
    if f'[{g.plan_name}]' not in txt:        
        with open(p / '_sidebar.md', 'a',encoding='UTF-8') as f:
            f.write(f'\n	* [{g.plan_name}](latest/{g.plan_name})')

    files = []
    for f in report.iterdir():
        if f.stem not in ['_sidebar', 'README']:
            files.append(f.stem)
    files.sort(reverse=True)
    with open(report / '_sidebar.md', 'w',encoding='UTF-8') as f:
        txt = f'* 「{g.plan_name}」测试结果\n'
        for stem in files: 
            txt +=f'\n    * [{stem}]({g.plan_name}/{stem})'
        f.write(txt)




def escape(data, well='#'):
    return data.replace('|', '\|').replace('<', '\<').replace('>', '\>').replace('\n', '<br>').replace('#', well)

def tm(stamp, dot=' '):
    if dot == ' ':
        return arrow.get(stamp/1000).to('local').format(f'YYYY-MM-DD{dot}HH:mm:ss').replace(':', '&#58;')
    elif dot == '_':
        return arrow.get(stamp/1000).to('local').format(f'YYYYMMDD{dot}HH:mm:ss').replace(':', '')


import time
from copy import deepcopy
from qingtest.globals import g, timestamp
from qingtest.windows import w
from qingtest.testcase import TestCase
from qingtest.log import logger


class TestSuite:
    def __init__(self, testsuite, sheet_name, report, conditions={}):
        self.testsuite = testsuite
        self.sheet_name = sheet_name
        self.report = report
        self.conditions = conditions
        self.result = {}

        # base 在整个测试套件中首先执行一次
        self.base_testcase = {}
        # setup 在每个测试用例执行之前执行一次
        self.setup_testcase = {}
        for testcase in self.testsuite:
            if testcase['condition'].lower() == 'base':
                self.base_testcase = testcase
            elif testcase['condition'].lower() == 'setup':
                self.setup_testcase = testcase
                testcase['flag'] = 'N'  # setup 用例只在执行其他普通用例之前执行
            elif testcase['condition'].lower() == 'snippet':
                g.snippet[testcase['id']] = testcase
                testcase['flag'] = 'N'
            elif testcase.get('set'):
                testcase['flag'] = 'N'
                if testcase['set'] not in g.caseset:
                    g.caseset[testcase['set']] = [testcase]
                else:
                    g.caseset[testcase['set']].append(testcase)

    def testsuite_start(self):
        self.result['no'] = g.no
        g.no += 1
        self.result['testsuite'] = self.sheet_name
        self.result['start_timestamp'] = timestamp()

    def testsuite_end(self):
        self.result['end_timestamp'] = timestamp()
        g.testsuite_data[self.sheet_name] = self.result

    def setup(self, testcase, case):       
        if self.setup_testcase:
            logger.info('*** SETUP testcase ↓ ***')
            logger.info('-'*50)
        else:
            logger.info('...No SETUP testcase need to run...')

        def run_setup(testcase):
            if testcase:
                tc = TestCase(testcase)
                tc.run()
                if testcase['result'] == 'success':
                    flag = 'Y'
                else:
                    flag = 'N'
            else:
                flag = 'O'
            return flag
        setup_flag = run_setup(deepcopy(self.setup_testcase))

        if setup_flag == 'N':
            base_flag = run_setup(deepcopy(self.base_testcase))
            if base_flag == 'Y':
                setup_flag = run_setup(deepcopy(self.setup_testcase))
                if setup_flag == 'N':
                    testcase['result'] = 'blocked'
                    case.block('Blocked', 'SETUP is not success')
                    logger.info('-'*50)
                    logger.info(f'>>> Run the testcase: {testcase["id"]}|{testcase["title"]}')
                    logger.warn('>>>>>>>>>>>>>>>>>>>> blocked <<<<<<<<<<<<<<<<<<<< SETUP is not success')
                    return False
            elif base_flag == 'O':
                testcase['result'] = 'blocked'
                case.block('Blocked', 'SETUP is not success')
                logger.info('-'*50)
                logger.info(f'>>> Run the testcase: {testcase["id"]}|{testcase["title"]}')
                logger.warn('>>>>>>>>>>>>>>>>>>>> blocked <<<<<<<<<<<<<<<<<<<< SETUP is not success')
                return False

        return True

    def run_testcase(self, testcase):
        # 根据筛选条件，把不需要执行的测试用例跳过
        flag = False
        for k, v in self.conditions.items():
            if not isinstance(v, list):
                v = [v]
            if testcase[k] not in v:
                testcase['result'] = 'skipped'
                flag = True
        if flag:
            return

        if testcase['condition'].lower() == 'base':
            logger.info('*** BASE testcase ↓ ***')

        if testcase['condition'].lower() == 'setup':
            return

        # 统计开始时间
        testcase['start_timestamp'] = timestamp()
        # xml 测试报告-测试用例初始化
        if testcase['flag'] != 'N':
            # 如果前置条件失败了，直接设置为阻塞
            if self.blocked_flag:
                testcase['result'] = 'blocked'
                testcase['end_timestamp'] = timestamp()
                return

            case = self.report.create_case(
                testcase['title'], testcase['id'])
            case.start()
            case.priority = testcase['priority']
            # 用例上下文
            self.previous = self.current
            self.current = testcase
        else:
            testcase['result'] = 'skipped'
            # case.skip('Skip', 'Autotest Flag is N')
            # logger.info('Run the testcase: %s|%s skipped, because the flag=N or the condition=snippet' % (
            #     testcase['id'], testcase['title']))
            # 统计结束时间
            testcase['end_timestamp'] = timestamp()
            return

        if testcase['condition'].lower() not in ('base', 'setup'):
            if testcase['condition'].lower() == 'sub':
                if self.previous['result'] != 'success':
                    testcase['result'] = 'blocked'
                    case.block(
                        'Blocked', 'Main or pre Sub testcase is not success')
                    logger.info('-'*50)
                    logger.info(f'>>> Run the testcase: {testcase["id"]}|{testcase["title"]}')                        
                    logger.warn('>>>>>>>>>>>>>>>>>>>> blocked <<<<<<<<<<<<<<<<<<<< Main or pre Sub TestCase is not success')
                    # 统计结束时间
                    testcase['end_timestamp'] = timestamp()
                    return
            # 如果前置条件为 skip，则此用例不执行前置条件
            elif testcase['condition'].lower() == 'skip':
                pass
            else:
                result = self.setup(testcase, case)
                # if result == 'N':
                if not result:
                    # 统计结束时间
                    testcase['end_timestamp'] = timestamp()
                    return

        try:
            tc = TestCase(testcase)
            logger.info('-'*50)
            tc.run()

            # 统计结束时间
            testcase['end_timestamp'] = timestamp()

            if testcase['result'] == 'success':
                case.succeed()
            elif testcase['result'] == 'failure':
                case.fail('Failure', testcase['report'])
                if testcase['condition'].lower() == 'base':
                    logger.warn('Run the testcase: %s|%s Failure, BASE is not success. Break the AutoTest' % (
                        testcase['id'], testcase['title']))
                    self.blocked_flag = True
                    return
                if testcase['condition'].lower() == 'setup':
                    logger.warn('Run the testcase: %s|%s failure, SETUP is not success. Break the AutoTest' % (
                        testcase['id'], testcase['title']))
                    self.blocked_flag = True
                    return
        except Exception as exception:
            case.error('Error', 'Remark:%s |||Exception:%s' %
                        (testcase['remark'], exception))
            logger.exception('Run the testcase: %s|%s failure' %
                                (testcase['id'], testcase['title']))
            if testcase['condition'].lower() == 'base':
                logger.warn('Run the testcase: %s|%s error, BASE is not success. Break the AutoTest' % (
                    testcase['id'], testcase['title']))
                self.blocked_flag = True
                return
            if testcase['condition'].lower() == 'setup':
                logger.warn('Run the testcase: %s|%s error, SETUP is not success. Break the AutoTest' % (
                    testcase['id'], testcase['title']))
                self.blocked_flag = True
                return


    def run(self):

        self.testsuite_start()

        # 当前测试用例
        self.current = {'result': 'success'}
        # 上一个测试用例
        self.previous = {}

        # 前置条件执行失败标志，即未执行用例阻塞标志
        self.blocked_flag = False

        for testcase in self.testsuite:
            self.run_testcase(testcase)

        self.report.finish()
        self.testsuite += g.casesets   # 把用例组合执行结果添加到末尾

        # 2.清理环境
        try:
            if g.platform.lower() in ('desktop',):
                # w.close()
                g.driver.quit()
                logger.info('--- Quit th Driver: %s' % g.browserName)
        except:
            logger.exception('Clear the env is fail')

        self.testsuite_end()






from pathlib import Path
import sys
import json
import requests
from qingtest.data import testsuite_format, testsuite2data, testsuite2report
from qingtest.parse import parse
from qingtest.elements import e
from qingtest.globals import g
from qingtest.windows import w
from qingtest.testsuite import TestSuite
from qingtest.utility import Excel, get_record, mkdir
from qingtest.log import logger, set_log
from qingtest.junit import JUnit
from qingtest.report import summary, markdown
from qingtest.config import _testcase, _elements, _report
from requests_toolbelt import MultipartEncoder


class Autotest:

    def __init__(self, file_name, sheet_name, desired_caps={}, server_url=''):
        if desired_caps:
            self.desired_caps = desired_caps
        else:
            self.desired_caps = {
                'platformName': 'Desktop', 'browserName': 'Chrome'}
        self.server_url = server_url
        self.conditions = {}
        g.plan_name = file_name.split('-')[0]
        g.init(self.desired_caps, self.server_url)

        plan_path = Path('snapshot') / g.plan_name
        task_path = plan_path / g.start_time[1:]

        for p in ('JUnit', 'report', 'snapshot', plan_path, task_path, 'report/' + g.plan_name):
            mkdir(p)

        g.plan_data['log'] = set_log(logger, task_path)

        self.testcase_file = str(
            Path('testcase') / (file_name + '-' + _testcase + '.xlsx'))
        self.elements_file = str(
            Path('element') / (g.plan_name + '-' + _elements + '.xlsx'))
        self.report_xml = str(
            Path('JUnit') / (file_name + '-' + _report + g.start_time + '.xml'))
        self.testcase_workbook = Excel(self.testcase_file, 'r')
        self.sheet_names = self.testcase_workbook.get_sheet(sheet_name)
        self.report_excel = str(Path(
            'report') / g.plan_name / (file_name + '-' + _report + g.start_time + '.xlsx'))
        self.report_workbook = Excel(self.report_excel, 'w')
        self.report_filename = str(
            file_name + '-' + _report + g.start_time + '.xlsx')

        self.report_data = {}  # 测试报告详细数据

    def fliter(self, **kwargs):
        # 筛选要执行的测试用例
        self.conditions = kwargs

    def plan(self):
        self.code = 0  # 返回码
        # 1.解析配置文件
        try:
            e.get_elements(self.elements_file)
        except:
            logger.exception('*** Parse config file failure ***')
            self.code = -1
            sys.exit(self.code)

        self.junit = JUnit()
        self.junit_suite = {}

        # 2.逐个执行测试套件
        for sheet_name in self.sheet_names:
            g.sheet_name = sheet_name
            # xml 测试报告初始化
            self.junit_suite[sheet_name] = self.junit.create_suite(
                g.plan_name, sheet_name)
            self.junit_suite[sheet_name].start()

            self.run(sheet_name)

        self.plan_data = g.plan_end()
        self.testsuite_data = g.testsuite_data

        summary_data = summary(
            self.plan_data, self.testsuite_data, self.report_data, {})
        self.report_workbook.write(summary_data, '_Summary_')
        self.report_workbook.close()

        with open(self.report_xml, 'w', encoding='utf-8') as f:
            self.junit.write(f)

    def run(self, sheet_name):
        # 1.从 Excel 获取测试用例集
        try:
            data = self.testcase_workbook.read(sheet_name)
            testsuite = testsuite_format(data)
            # logger.info('Testsuite imported from Excel:\n' +
            #             json.dumps(testsuite, ensure_ascii=False, indent=4))
            logger.info('From Excel import testsuite success')
        except:
            logger.exception('*** From Excel import testsuite failure ***')
            self.code = -1
            sys.exit(self.code)

        # 2.初始化全局对象
        try:
            g.set_driver()
            # 如果测试数据文件存在，则从该文件里读取数据，赋值到全局变量列表里
            data_file = Path('data') / (g.plan_name +
                                        '-' + sheet_name + '.csv')
            if data_file.is_file():
                g.test_data = get_record(str(data_file))
            w.init()
        except:
            logger.exception('*** Init global object failure ***')
            self.code = -1
            sys.exit(self.code)

        # 3.解析测试用例集
        try:
            parse(testsuite)
            logger.debug('testsuite has been parsed:\n' + str(testsuite))
        except:
            logger.exception('*** Parse testsuite failure ***')
            self.code = -1
            sys.exit(self.code)

        # 4.执行测试套件
        g.ts = TestSuite(testsuite, sheet_name,
                         self.junit_suite[sheet_name], self.conditions)
        g.ts.run()

        # 5.判断测试结果
        if self.junit_suite[sheet_name].high_errors + self.junit_suite[sheet_name].medium_errors + \
                self.junit_suite[sheet_name].high_failures + self.junit_suite[sheet_name].medium_failures:
            self.code = -1

        # 6.保存测试结果
        try:
            data = testsuite2data(testsuite)
            self.report_workbook.write(data, sheet_name)
            self.report_data[sheet_name] = testsuite2report(testsuite)
        except:
            logger.exception('*** Save the report is failure ***')

    def report(self, md_path):
        markdown(self.plan_data, self.testsuite_data,
                 self.report_data, md_path)

    def sendMessage(key, status, name):
        headers = {'content-type': "application/json"}
        url=f"https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key={key}"
        if status == 1:
            result = "测试通过"
        else:
            result = "测试失败"

        content = {
            "msgtype": "text",
            "text": {
                "content": "自动化测试用例：\n" + "用例名："+name + "\n测试结果：" + result,
                "mentioned_mobile_list": ["18939881330"]
            }
        }
        requests.post(url=url, data=json.dumps(content), headers=headers)

    def sendFile(key,file_id):
        url=f"https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key={key}"
        headers = {'content-type': "application/json"}
        content = {
            "msgtype": "file",
            "file": {
                "media_id": file_id
                 }
            }
        s=requests.post(url=url, data=json.dumps(content), headers=headers)
        print(s.text)

    def uploadFile(filepath, filename, access_token):
        post_file_url = f"https://qyapi.weixin.qq.com/cgi-bin/webhook/upload_media?key={access_token}&type=file"

        # m = MultipartEncoder(fields:{filename: ('file', open(filepath + Autotest., 'rb'), 'text/plain')})

        m = MultipartEncoder(
        fields = {filename: (filename, open(filepath, 'rb'))},
    )
        print(m)
        r=requests.post(url = post_file_url, data = m, headers = {
                        'Content-Type': "multipart/form-data"})
        print(r.text)
        return r.json()['media_id']



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qingtest",
    version="0.0.4",
    author="zhoujiahui",
    author_email="zhoujiahui@exiao.tech",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.qingflow.com",
    packages=setuptools.find_packages(),
    install_requires=['selenium==3.141.0','xlrd==1.2.0','XlsxWriter==1.3.7','requests==2.24.0','Appium-Python-Client==1.0.2','Pillow==8.0.1','openpyxl==3.0.5','arrow==0.17.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from time import sleep
import re
from qingtest.globals import g
from qingtest.elements import e
from qingtest.windows import w
from qingtest.locator import locating_elements, locating_data, locating_element
from qingtest.log import logger
from qingtest.parse import data_format
from qingtest.utility import compare
from appium.webdriver.common.touch_action import TouchAction
from selenium.common.exceptions import ElementClickInterceptedException


class Common():
    @classmethod
    def title(cls, data, output):
        logger.info('DATA:%s' % repr(data['text']))
        logger.info('REAL:%s' % repr(g.driver.title))
        if data['text'].startswith('*'):
            assert data['text'][1:] in g.driver.title
        else:
            assert data['text'] == g.driver.title
        # 只能获取到元素标题
        for key in output:
            g.var[key] = g.driver.title

    @classmethod
    def current_url(cls, data, output):
        logger.info('DATA:%s' % repr(data['text']))
        logger.info('REAL:%s' % repr(g.driver.current_url))
        if data['text'].startswith('*'):
            assert data['text'][1:] in g.driver.current_url
        else:
            assert data['text'] == g.driver.current_url
        # 只能获取到元素 url
        for key in output:
            g.var[key] = g.driver.current_url


def check(step):
    data = step['data']
    if not data:
        data = step['expected']

    element = step['element']
    element_location = locating_element(element)
    if '#' in element:
        e_name = element.split('#')[0] + '#'
    else:
        e_name = element
    by = e.elements[e_name]['by']
    output = step['output']

    if by in ('title', 'current_url'):
        getattr(Common, by)(data, output)

    else:
        for key in data:
            # 预期结果
            expected = data[key]
            # 切片操作处理
            s = re.findall(r'\[.*?\]', key)
            if s:
                s = s[0]
                key = key.replace(s, '')

            if key == 'text':
                real = element_location.text
            else:
                real = element_location.get_attribute(key)
            if s:
                real = eval('real' + s)

            logger.info('DATA:%s' % repr(expected))
            logger.info('REAL:%s' % repr(real))
            compare(expected, real)

        # 获取元素其他属性
        for key in output:
            if output[key] == 'text':
                g.var[key] = element_location.text
            elif output[key] in ('text…', 'text...'):
                if element_location.text.endswith('...'):
                    g.var[key] = element_location.text[:-3]
                else:
                    g.var[key] = element_location.text
            else:
                g.var[key] = element_location.get_attribute(output[key])


def notcheck(step):
    data = step['data']
    if not data:
        data = step['expected']

    element = step['element']
    # element_location = locating_element(element)

    if e.elements[element]['by'] == 'title':
        assert data['text'] != g.driver.title


def input(step):
    data = step['data']
    element = step['element']
    element_location = locating_element(element)

    if isinstance(data['text'], tuple):
        element_location.send_keys(*data['text'])
    elif element_location:
        if step['data'].get('清除文本', '') == '否' or step['data'].get('clear', '').lower() == 'no':
            pass
        else:
            element_location.clear()
        element_location.send_keys(data['text'])


def set_value(step):
    data = step['data']
    element = step['element']
    element_location = locating_element(element)

    if isinstance(data['text'], tuple):
        element_location.set_value(*data['text'])
    elif element_location:
        if step['data'].get('清除文本', '') == '否' or step['data'].get('clear', '').lower() == 'no':
            pass
        else:
            element_location.clear()
        element_location.set_value(data['text'])


def click(step):
    element = step['element']
    if isinstance(element, str):
        #element_location = locating_element(element, 'CLICK')
        element_location = locating_element(element)
        try:
            element_location.click()
        except ElementClickInterceptedException:  # 如果元素为不可点击状态，则等待1秒，再重试一次
            sleep(1)
            element_location.click()
    elif isinstance(element, list):
        for _e in element:
            #element_location = locating_element(_e, 'CLICK')
            element_location = locating_element(_e)
            try:
                element_location.click()
            except ElementClickInterceptedException:  # 如果元素为不可点击状态，则等待1秒，再重试一次
                sleep(1)
                element_location.click()
            sleep(0.5)
    sleep(0.5)

    # 获取元素其他属性
    output = step['output']
    for key in output:
        if output[key] == 'text':
            g.var[key] = element_location.text
        elif output[key] == 'tag_name':
            g.var[key] = element_location.tag_name
        elif output[key] in ('text…', 'text...'):
            if element_location.text.endswith('...'):
                g.var[key] = element_location.text[:-3]
            else:
                g.var[key] = element_location.text
        else:
            g.var[key] = element_location.get_attribute(output[key])

    # if w.current_context.startswith('WEBVIEW'):
    #     # 判断是否打开了新的窗口，并将新窗口添加到所有窗口列表里
    #     all_handles = g.driver.window_handles
    #     for handle in all_handles:
    #         if handle not in w.windows.values():
    #             w.register(step, handle)


def tap(step):
    action = TouchAction(g.driver)

    element = step['element']
    if isinstance(element, str):

        if ',' in element:
            position = element.split(',')
            x = int(position[0])
            y = int(position[1])
            position = (x, y)
            g.driver.tap([position])
        else:
            element_location = locating_element(element, 'CLICK')
            action.tap(element_location).perform()
    elif isinstance(element, list):
        if ',' in element[0]:
            for el in element:
                position = el.split(',')
                x = int(position[0])
                y = int(position[1])
                position = (x, y)
                g.driver.tap([position])
                sleep(0.5)
        else:
            for _e in element:
                element_location = locating_element(_e, 'CLICK')
                action.tap(element_location).perform()
                sleep(0.5)
    sleep(0.5)

    # 获取元素其他属性
    output = step['output']
    for key in output:
        if output[key] == 'text':
            g.var[key] = element_location.text
        elif output[key] == 'tag_name':
            g.var[key] = element_location.tag_name
        elif output[key] in ('text…', 'text...'):
            if element_location.text.endswith('...'):
                g.var[key] = element_location.text[:-3]
            else:
                g.var[key] = element_location.text
        else:
            g.var[key] = element_location.get_attribute(output[key])

    # if w.current_context.startswith('WEBVIEW'):
    #     # 判断是否打开了新的窗口，并将新窗口添加到所有窗口列表里
    #     all_handles = g.driver.window_handles
    #     for handle in all_handles:
    #         if handle not in w.windows.values():
    #             w.register(step, handle)


def press_keycode(step):
    element = step['element']
    g.driver.press_keycode(int(element))


def swipe(step):
    element = step['element']
    duration = step['data'].get('持续时间', 0.3)
    assert isinstance(element, list) and len(
        element) == 2, '坐标格式或数量不对，正确格式如：100,200|300,400'

    start = element[0].replace('，', ',').split(',')
    start_x = int(start[0])
    start_y = int(start[1])

    end = element[1].replace('，', ',').split(',')
    end_x = int(end[0])
    end_y = int(end[1])

    if duration:
        g.driver.swipe(start_x, start_y, end_x, end_y, sleep(float(duration)))
    else:
        g.driver.swipe(start_x, start_y, end_x, end_y)


def line(step):
    element = step['element']
    duration = float(step['data'].get('持续时间', 0.3))
    assert isinstance(element, list) and len(
        element) > 1, '坐标格式或数量不对，正确格式如：258,756|540,1032'
    postions = []
    for _e in element:
        _e = _e.replace('，', ',')
        p = _e.split(',')
        postions.append(p)

    action = TouchAction(g.driver)
    action = action.press(
        x=postions[0][0], y=postions[0][1]).wait(duration * 1000)
    for i in range(1, len(postions)):
        action.move_to(x=postions[i][0], y=postions[i]
                       [1]).wait(duration * 1000)
    action.release().perform()


def line_unlock(step):
    element = step['element']
    duration = float(step['data'].get('持续时间', 0.3))
    assert isinstance(element, list) and len(
        element) > 2, '坐标格式或数量不对，正确格式如：lock_pattern|1|4|7|8|9'
    _e = locating_element(element[0])
    rect = _e.rect
    w = rect['width'] / 6
    h = rect['height'] / 6

    key = {}
    key['1'] = (rect['x'] + 1 * w, rect['y'] + 1 * h)
    key['2'] = (rect['x'] + 3 * w, rect['y'] + 1 * h)
    key['3'] = (rect['x'] + 5 * w, rect['y'] + 1 * h)
    key['4'] = (rect['x'] + 1 * w, rect['y'] + 3 * h)
    key['5'] = (rect['x'] + 3 * w, rect['y'] + 3 * h)
    key['6'] = (rect['x'] + 5 * w, rect['y'] + 3 * h)
    key['7'] = (rect['x'] + 1 * w, rect['y'] + 5 * h)
    key['8'] = (rect['x'] + 3 * w, rect['y'] + 5 * h)
    key['9'] = (rect['x'] + 5 * w, rect['y'] + 5 * h)

    action = TouchAction(g.driver)
    for i in range(1, len(element)):
        k = element[i]
        if i == 1:
            action = action.press(
                x=key[k][0], y=key[k][1]).wait(duration * 1000)
        action.move_to(x=key[k][0], y=key[k][1]).wait(duration * 1000)
    action.release().perform()


def rocker(step):
    element = step['element']
    duration = float(step['data'].get('持续时间', 0.3))
    rocker_name = step['data'].get('摇杆', 'rocker')
    release = step['data'].get('释放', False)

    if isinstance(element, str):
        if element:
            element = [element]
        else:
            element = []

    postions = []
    for _e in element:
        _e = _e.replace('，', ',')
        p = _e.split(',')
        postions.append(p)

    # 如果 action 中么有此摇杆名，则是新的遥感
    if not g.action.get(rocker_name):
        g.action[rocker_name] = TouchAction(g.driver)
        g.action[rocker_name].press(
            x=postions[0][0], y=postions[0][1]).wait(duration * 1000)
        # 新摇杆的第一个点已操作，需要删除
        postions.pop(0)
    # 依次操作
    for i in range(len(postions)):
        g.action[rocker_name].move_to(
            x=postions[i][0], y=postions[i][1]).wait(duration * 1000)

    if release:
        # 释放摇杆，并删除摇杆
        g.action[rocker_name].release().perform()
        del g.action[rocker_name]
    else:
        g.action[rocker_name].perform()


def scroll(step):
    element = step['element']
    assert isinstance(element, list) and len(
        element) == 2, '元素格式或数量不对，正确格式如：origin_el|destination_el'
    origin = locating_element(element[0])
    destination = locating_element(element[1])
    g.driver.scroll(origin, destination)


def flick_element(step):
    element = step['element']
    speed = step['data'].get('持续时间', 10)
    assert isinstance(element, list) and len(
        element) == 2, '坐标格式或数量不对，正确格式如：elment|200,300'
    _e = eval(element[0])

    end = element[1].replace('，', ',').split(',')
    end_x = int(end[0])
    end_y = int(end[1])

    if speed:
        g.driver.flick_element(_e, end_x, end_y, int(speed))


def flick(step):
    element = step['element']
    assert isinstance(element, list) and len(
        element) == 2, '坐标格式或数量不对，正确格式如：100,200|300,400'

    start = element[0].replace('，', ',').split(',')
    start_x = int(start[0])
    start_y = int(start[1])

    end = element[1].replace('，', ',').split(',')
    end_x = int(end[0])
    end_y = int(end[1])

    g.driver.flick(start_x, start_y, end_x, end_y)


def drag_and_drop(step):
    element = step['element']
    assert isinstance(element, list) and len(
        element) == 2, '元素格式或数量不对，正确格式如：origin_el|destination_el'
    origin = locating_element(element[0])
    destination = locating_element(element[1])
    g.driver.drag_and_drop(origin, destination)


def long_press(step):
    action = TouchAction(g.driver)

    element = step['element']
    duration = step['data'].get('持续时间', 1000)
    if ',' in element or '，' in element:
        position = element.replace('，', ',').split(',')
        x = int(position[0])
        y = int(position[1])
        action.long_press(x=x, y=y, duration=duration).perform()
    else:
        element_location = locating_element(element)
        action.long_press(element_location, duration=duration).perform()
    sleep(0.5)


def pinch(step):
    element = step['element']
    element_location = locating_element(element[0])
    percent = step['data'].get('百分比', 200)
    steps = step['data'].get('步长', 50)
    g.driver.pinch(element_location, percent, steps)


def zoom(step):
    element = step['element']
    element_location = locating_element(element[0])
    percent = step['data'].get('百分比', 200)
    steps = step['data'].get('步长', 50)
    g.driver.zoom(element_location, percent, steps)


def hide_keyboard(step):
    g.driver.hide_keyboard()


def shake(step):
    g.driver.shake()


def launch_app(step):
    g.driver.launch_app()


def is_locked(step):
    status = g.driver.is_locked()
    assert status, "it's not locked"


def lock(step):
    g.driver.lock()


def unlock(step):
    g.driver.unlock()


def tab_name(step):
    element = step['element']
    name = step['data']['text']
    # 从所有窗口中查找给定元素，如果查询到就命名，否则报错
    all_handles = g.driver.window_handles
    logger.info('All Handles: %s' % all_handles)

    flag = False
    for handle in all_handles:
        #logger.info('Page Source: %s \n%s' % (handle, g.driver.page_source))
        #logger.info('All Windows: %s' %w.windows)
        if handle not in w.windows.values():
            # 切换至此窗口
            g.driver.switch_to_window(handle)
            try:
                # 成功定位到关键元素
                locating_element(element, 'CLICK')
                # 添加到窗口资源池 g.windows
                w.windows[name] = handle
                # 把当前窗口名字改为新窗口名称
                w.current_window = name
                flag = True
                logger.info('Current Window: %s' % repr(name))
                logger.info('Current Handle: %s' % repr(handle))
            except:
                pass
    if not flag:
        raise Exception(
            'Tab Name failure: the element:%s in all tab is not found' % element)




from pathlib import Path
from PIL import Image
from PIL import ImageChops
import math
import operator
import time
from functools import reduce
from qingtest.globals import g, now
from qingtest.log import logger
from qingtest.utility import mkdir


# 解决打开图片大于 20M 的限制
Image.MAX_IMAGE_PIXELS = None

def crop(element, src, target):
    location = element.location
    size = element.size
    im = Image.open(src)
    left = location['x']
    top = location['y']
    right = location['x'] + size['width']
    bottom = location['y'] + size['height']
    im = im.crop((left, top, right, bottom))
    im.save(target)


def blank(src, boxs):
    white = Image.new('RGB',(5000, 5000), 'white')   
    im = Image.open(src)
    for box in boxs:
        w = white.crop(box)
        im.paste(w, box[:2])
    im.save(src)


def cut(src, target, box):
    im = Image.open(src)
    im = im.crop(box)
    im.save(target)


def get_screenshot(file_path):
    if g.headless:
        width = g.driver.execute_script("return Math.max(document.body.scrollWidth, document.documentElement.clientWidth, document.documentElement.scrollWidth, document.documentElement.offsetWidth);")
        height = g.driver.execute_script("return Math.max(document.body.scrollHeight, document.documentElement.clientHeight, document.documentElement.scrollHeight, document.documentElement.offsetHeight);")
        g.driver.set_window_size(width,height)
        time.sleep(3)
    g.driver.get_screenshot_as_file(file_path)
    if g.headless:
        g.driver.set_window_size(1920,1080)


class Snapshot:
    def __init__(self):
        snapshot_plan = Path('snapshot') / g.plan_name
        self.snapshot_folder = snapshot_plan / g.start_time[1:]
        snapshot_expected = Path('snapshot') / 'expected'
        self.expected_folder = snapshot_expected / g.plan_name
        for p in (snapshot_plan, self.snapshot_folder, snapshot_expected, self.expected_folder):
            mkdir(p)

    def pre(self, step, label):
        self.label = label
        self.screen_flag = False
        self.element_flag = False

        # 处理输出数据中的截图设置
        self.output = {}
        for k, v in dict(step['output'].items()).items():
            if v in ('#screen_shot','#ScreenShot'):
                self.output['#screen_shot'] = k
                step['output'].pop(k)
                self.screen_flag = True
            if v == ('#element_shot', '#ElementShot'):
                self.output['#element_shot'] = k
                step['output'].pop(k)
                self.element_flag = True

        # 处理图片比较
        self.expected = {}
        for data in (step['data'], step['expected']):
            if '#screen_shot' in data:
                self.screen_flag = True
                p = Path(data['#screen_shot']).stem
                self.expected['#screen_name'] = '('+p.split('[')[1].split(']')[0]+')' if '[' in p else p
                if Path(data['#screen_shot']).is_file():
                                     
                    step['snapshot']['expected_screen'] = data.pop('#screen_shot')
                else:                  
                    step['snapshot']['expected_screen'] = str(self.expected_folder / data.pop('#screen_shot'))

            if '#element_shot' in data:
                self.element_flag = True
                p = Path(data['#element_shot']).stem
                self.expected['#element_name'] = '('+p.split('[')[1].split(']')[0]+')' if '[' in p else p
                if Path(data['#element_shot']).is_file():
                    step['snapshot']['expected_element'] = data.pop('#element_shot')
                else:
                    step['snapshot']['expected_element'] = str(self.expected_folder / data.pop('#element_shot'))

            if '#cut' in data:
                self.expected['#cut'] = data.pop('#cut')
            if '#blank' in data:
                self.expected['#blank'] = data.pop('#blank')

    def web_screen(self, step, element):
        # 截图
        screen_v = self.output.get('#screen_shot', '')
        element_v = self.output.get('#element_shot', '')

        if g.snapshot or self.screen_flag or self.element_flag:
            from selenium.webdriver.support import expected_conditions as EC
            if not EC.alert_is_present()(g.driver):
                if self.expected.get('#screen_name'):
                    screen_name = self.expected['#screen_name']
                elif screen_v:
                    screen_name = screen_v
                else:
                    screen_name = ''
                if screen_name:
                    file_name = self.label + now() + '#screen' + '[' + screen_name + ']' + '.png'
                else:
                    file_name = self.label + now() + '#screen' + '.png'                                        
                step['snapshot']['real_screen'] = str(
                    Path(self.snapshot_folder) / file_name)
                get_screenshot(step['snapshot']['real_screen'])
                if screen_v:
                    g.var[screen_v] = step['snapshot']['real_screen']

        if element_v:
            file_name = self.label + now() + '#element' + '[' + element_v + ']' + '.png'
            step['snapshot']['real_element'] = str(Path(self.snapshot_folder) / file_name)
            crop(element, step['snapshot']['real_screen'], step['snapshot']['real_element'])
            g.var[element_v] = step['snapshot']['real_element']

    def web_check(self, step, element):

        def deep(src):
            # 把不需要比较的部分贴白
            if self.expected.get('#blank'):
                blank(src, eval(self.expected.get('#blank')))
            # 裁剪需要比较的部分
            if self.expected.get('#cut'):
                cut(src, src, eval(self.expected.get('#cut')))

        if Path(step['snapshot'].get('expected_screen', '')).is_file():
            # 屏幕截图比较
            image1 = Image.open(step['snapshot']['expected_screen'])
            image2 = Image.open(step['snapshot']['real_screen'])
            deep(step['snapshot']['real_screen'])
            histogram1 = image1.histogram()
            histogram2 = image2.histogram()
            differ = math.sqrt(reduce(operator.add, list(
                map(lambda a, b: (a - b)**2, histogram1, histogram2))) / len(histogram1))
            diff = ImageChops.difference(image1.convert('RGB'), image2.convert('RGB'))
            if differ < 0.1:
                # 图片间没有任何不同
                logger.info('SnapShot: screen_shot is the same')
            else:
                file_name = self.label + now() + 'diff_screen' + '.png'
                step['snapshot']['diff_screen'] = str(
                    Path(self.snapshot_folder) / file_name)
                diff.save(step['snapshot']['diff_screen'])
                raise Exception('SnapShot: screen_shot is diff: %s' % differ)
        elif step['snapshot'].get('expected_screen'):
            get_screenshot(step['snapshot']['expected_screen'])
            deep(step['snapshot']['expected_screen'])

        if Path(step['snapshot'].get('expected_element', '')).is_file():
            file_name = self.label + now() + '#element' + '[' + self.expected['#element_name'] + ']' + '.png'
            step['snapshot']['real_element'] = str(Path(self.snapshot_folder) / file_name)
            crop(element, step['snapshot']['real_screen'], step['snapshot']['real_element'])
            deep(step['snapshot']['real_element'])

            # 屏幕截图比较
            image1 = Image.open(step['snapshot']['expected_element'])
            image2 = Image.open(step['snapshot']['real_element'])
            histogram1 = image1.histogram()
            histogram2 = image2.histogram()
            differ = math.sqrt(reduce(operator.add, list(
                map(lambda a, b: (a - b)**2, histogram1, histogram2))) / len(histogram1))
            diff = ImageChops.difference(image1.convert('RGB'), image2.convert('RGB'))
            if differ < 0.1:
                logger.info('SnapShot: element_shot is the same')
            else:
                file_name = self.label + now() + 'diff_element' + '.png'
                step['snapshot']['diff_element'] = str(
                    Path(self.snapshot_folder) / file_name)
                diff.save(step['snapshot']['diff_element'])
                raise Exception('SnapShot: element_shot is diff: %s' % differ)
        elif step['snapshot'].get('expected_element'):
            crop(element, step['snapshot']['real_screen'], step['snapshot']['expected_element'])
            deep(step['snapshot']['expected_element'])

    def web_shot(self, step, element):
        self.web_screen(step, element)
        self.web_check(step, element)


    def windwos_capture(self, dialog, step):
        # 截图
        screen_v = self.output.get('#screen_shot', '')
        element_v = self.output.get('#element_shot', '')

        if g.snapshot or self.screen_flag:
            if self.expected.get('#screen_name'):
                screen_name = self.expected['#screen_name']
            elif screen_v:
                screen_name = screen_v
            else:
                screen_name = ''
            if screen_name:
                file_name = self.label + now() + '#screen' + '[' + screen_name + ']' + '.png'
            else:
                file_name = self.label + now() + '#screen' + '.png'               
            step['snapshot']['real_screen'] = str(Path(self.snapshot_folder) / file_name)
            pic = dialog.capture_as_image()
            pic.save(step['snapshot']['real_screen'])	
            if screen_v:
                g.var[screen_v] = step['snapshot']['real_screen']

        if element_v:
            file_name = self.label + now() + '#element' + '[' + element_v + ']' + '.png'
            step['snapshot']['real_element'] = str(Path(self.snapshot_folder) / file_name)

            element = step['element']
            if dialog.backend.name == 'win32':
                pic = dialog.window(best_match=element).capture_as_image()
            elif dialog.backend.name == 'uia':
                pic = dialog.child_window(best_match=element).capture_as_image()       
            pic.save(step['snapshot']['real_screen'])	            
            g.var[element_v] = step['snapshot']['real_element']


    def windwos_check(self, dialog, step):
        element = step['element']        
        if Path(step['snapshot'].get('expected_screen', '')).is_file():
            # 屏幕截图比较
            image1 = Image.open(step['snapshot']['expected_screen'])
            image2 = Image.open(step['snapshot']['real_screen'])
            histogram1 = image1.histogram()
            histogram2 = image2.histogram()
            differ = math.sqrt(reduce(operator.add, list(
                map(lambda a, b: (a - b)**2, histogram1, histogram2))) / len(histogram1))
            diff = ImageChops.difference(image1.convert('RGB'), image2.convert('RGB'))
            if differ < 0.1:
                # 图片间没有任何不同
                logger.info('SnapShot: screen_shot is the same')
            else:
                file_name = self.label + now() + 'diff_screen' + '.png'
                step['snapshot']['diff_screen'] = str(
                    Path(self.snapshot_folder) / file_name)
                diff.save(step['snapshot']['diff_screen'])
                raise Exception('SnapShot: screen_shot is diff: %s' % differ)
        elif step['snapshot'].get('expected_screen'):
            pic = dialog.capture_as_image()
            pic.save(step['snapshot']['expected_screen'])	            

        if Path(step['snapshot'].get('expected_element', '')).is_file():
            file_name = self.label + now() + '#element' + '.png'
            step['snapshot']['real_element'] = str(Path(self.snapshot_folder) / file_name)
            if dialog.backend.name == 'win32':
                pic = dialog.window(best_match=element).capture_as_image()
            elif dialog.backend.name == 'uia':
                pic = dialog.child_window(best_match=element).capture_as_image() 
            pic.save(step['snapshot']['real_element'])	            

            # 屏幕截图比较
            image1 = Image.open(step['snapshot']['expected_element'])
            image2 = Image.open(step['snapshot']['real_element'])
            histogram1 = image1.histogram()
            histogram2 = image2.histogram()
            differ = math.sqrt(reduce(operator.add, list(
                map(lambda a, b: (a - b)**2, histogram1, histogram2))) / len(histogram1))
            diff = ImageChops.difference(image1.convert('RGB'), image2.convert('RGB'))
            if differ < 0.1:
                logger.info('SnapShot: element_shot is the same')
            else:
                file_name = self.label + now() + 'diff_element' + '.png'
                step['snapshot']['diff_element'] = str(
                    Path(self.snapshot_folder) / file_name)
                diff.save(step['snapshot']['diff_element'])
                raise Exception('SnapShot: element_shot is diff: %s' % differ)
        elif step['snapshot'].get('expected_element'):
            if dialog.backend.name == 'win32':
                pic = dialog.window(best_match=element).capture_as_image()
            elif dialog.backend.name == 'uia':
                pic = dialog.child_window(best_match=element).capture_as_image()       
            pic.save(step['snapshot']['expected_element'])	            


    def windows_shot(self, dialog, step):
        self.windwos_capture(dialog, step)
        self.windwos_check(dialog, step)